
# This file was generated by 'versioneer.py' (0.20) from
# revision-control system data, or from the parent directory name of an
# unpacked source archive. Distribution tarballs contain a pre-generated copy
# of this file.

import json

version_json = '''
{
 "date": "2025-06-25T13:33:29+0200",
 "dirty": false,
 "error": null,
 "full-revisionid": "519a7c4f0ef6c7a3d5550c4018a2c01bbb060d3e",
 "version": "2.2"
}
'''  # END VERSION_JSON


def get_versions():
    return json.loads(version_json)
